'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _cloneDeep = require('clone-deep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

var _kindOf = require('kind-of');

var _kindOf2 = _interopRequireDefault(_kindOf);

var _defaultTypes = require('./default-types');

var _defaultTypes2 = _interopRequireDefault(_defaultTypes);

var _structError = require('./struct-error');

var _structError2 = _interopRequireDefault(_structError);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * A private string to identify structs by.
 *
 * @type {String}
 */

const IS_STRUCT = '@@__STRUCT__@@';

/**
 * Public methods for struct functions.
 *
 * @type {Array}
 */

const STRUCT_METHODS = ['assert', 'default', 'defaultProperty', 'mask', 'test', 'validate', 'validateElement', 'validateProperty'];

/**
 * The diffrent Kinds of struct schemas.
 *
 * @type {Array}
 */

const STRUCT_KINDS = ['scalar', 'function', 'object', 'list'];

/**
 * Convenience flags for the struct factory.
 *
 * @type {Array}
 */

const STRUCT_FLAGS = ['required', 'unsealed'];

/**
 * The error codes that structs can throw.
 *
 * @type {String}
 */

const ELEMENT_INVALID = 'element_invalid';
const PROPERTY_INVALID = 'property_invalid';
const PROPERTY_REQUIRED = 'property_required';
const PROPERTY_UNKNOWN = 'property_unknown';
const VALUE_INVALID = 'value_invalid';
const VALUE_REQUIRED = 'value_required';

/**
 * Create a struct factory with a `config`.
 *
 * @param {Object} config
 * @return {Function}
 */

function superstruct(config = {}) {
  const TYPES = _extends({}, _defaultTypes2.default, config.types || {});

  /**
   * Base schema.
   *
   * @type {Schema}
   */

  class Schema {

    /**
     * Create a schema with `schema`, `defaults` and `options`.
     *
     * @param {Any} schema
     * @param {Any} defaults
     * @param {Object} options
     */

    constructor(schema, defaults, options = {}) {
      this.schema = schema;
      this.defaults = defaults;
      this.options = options;
      this.type = null;
    }

    /**
     * Assert that a `value` is valid, throwing if not.
     *
     * @param {Any} value
     * @return {Any|StructError}
     */

    assert(value) {
      const result = this.validate(value);

      if (result instanceof _structError2.default) {
        throw result;
      }

      return result;
    }

    /**
     * Get the defaulted value, given an initial `value`.
     *
     * @param {Any} value
     * @return {Any}
     */

    default(value) {
      if (value !== undefined) return value;
      const { defaults } = this;
      return typeof defaults === 'function' ? defaults() : (0, _cloneDeep2.default)(defaults);
    }

    /**
     * Test that a `value` is valid.
     *
     * @param {Any} value
     * @return {Boolean}
     */

    test(value) {
      const result = this.validate(value);
      return !(result instanceof _structError2.default);
    }

    /**
     * Validate a `value`, returning an error or the value with defaults.
     *
     * @param {Any} value
     * @return {Any|StructError}
     */

    validate(value) {
      value = this.default(value);
      return value;
    }

  }

  /**
   * Function schema.
   *
   * @type {FunctionSchema}
   */

  class FunctionSchema extends Schema {

    /**
     * Validate a `value`, returning an error or the value with defaults.
     *
     * @param {Any} value
     * @return {Any|StructError}
     */

    validate(value) {
      const { options, type, schema } = this;

      value = this.default(value);

      if (options.required && value === undefined) {
        return new _structError2.default(VALUE_REQUIRED, { type });
      }

      if (value !== undefined && !schema(value)) {
        return new _structError2.default(VALUE_INVALID, { type, value });
      }

      return value;
    }

  }

  /**
   * Scalar schema.
   *
   * @type {ScalarSchema}
   */

  class ScalarSchema extends Schema {

    /**
     * Create a scalar schema with `schema`, `defaults` and `options`.
     *
     * @param {Any} schema
     * @param {Any} defaults
     * @param {Object} options
     */

    constructor(schema, defaults, options = {}) {
      super(schema, defaults, options);

      const required = !schema.endsWith('?');
      const type = required ? schema : schema.slice(0, -1);
      const types = type.split(/\s*\|\s*/g);
      const validators = types.map(t => {
        const fn = TYPES[t];
        if (typeof fn === 'function') return fn;
        throw new Error(`No struct validator function found for type "${t}".`);
      });

      this.options.required = required;
      this.type = type;
      this.validators = validators;
    }

    /**
     * Validate a `value`, returning an error or the value with defaults.
     *
     * @param {Any} value
     * @return {Any|StructError}
     */

    validate(value) {
      const { options, type, validators } = this;

      value = this.default(value);

      if (options.required && value === undefined) {
        return new _structError2.default(VALUE_REQUIRED, { type });
      }

      if (value !== undefined && !validators.some(fn => fn(value))) {
        return new _structError2.default(VALUE_INVALID, { type, value });
      }

      return value;
    }

  }

  /**
   * List schema.
   *
   * @type {ListSchema}
   */

  class ListSchema extends Schema {

    /**
     * Create a list schema with `schema`, `defaults` and `options`.
     *
     * @param {Any} schema
     * @param {Any} defaults
     * @param {Object} options
     */

    constructor(schema, defaults, options = {}) {
      super(schema, defaults, options);

      if (schema.length !== 1) {
        throw new Error(`List structs must be defined as an array with a single element, but you passed ${schema.length} elements.`);
      }

      const type = options.required ? 'array' : 'array?';
      const valueStruct = struct(type);
      const elementStruct = struct(schema[0]);

      this.type = type;
      this.valueStruct = valueStruct;
      this.elementStruct = elementStruct;
    }

    /**
     * Validate a list `element` at `index`.
     *
     * @param {Number} index
     * @param {Any} element
     * @return {Any|StructError}
     */

    validateElement(index, element) {
      const s = this.elementStruct;
      const result = s.validate(element);

      if (result instanceof _structError2.default) {
        const e = result;
        const path = [index].concat(e.path);

        if (e.code === VALUE_INVALID) {
          return new _structError2.default(ELEMENT_INVALID, _extends({}, e, { index, path }));
        }

        if ('path' in e) e.path = path;
        return e;
      }

      return result;
    }

    /**
     * Validate a list `value`.
     *
     * @param {Any} value
     * @return {Any|StructError}
     */

    validate(value) {
      const { options, valueStruct } = this;

      value = this.default(value);
      const result = valueStruct.validate(value);

      if (result instanceof _structError2.default) {
        return result;
      }

      const errors = [];
      const values = [];
      const isUndefined = !options.required && value === undefined;
      value = isUndefined ? [] : value;

      value.forEach((e, i) => {
        const r = this.validateElement(i, e);

        if (r instanceof _structError2.default) {
          errors.push(r);
        } else {
          values.push(r);
        }
      });

      if (errors.length) {
        const first = errors[0];
        first.errors = errors;
        return first;
      }

      return isUndefined && !values.length ? undefined : values;
    }

  }

  /**
   * Object schema.
   *
   * @type {ObjectSchema}
   */

  class ObjectSchema extends Schema {

    /**
     * Create an object schema with `schema`, `defaults` and `options`.
     *
     * @param {Any} schema
     * @param {Any} defaults
     * @param {Object} options
     */

    constructor(schema, defaults, options = {}) {
      super(schema, defaults, options);

      const type = options.required ? 'object' : 'object?';
      const valueStruct = struct(type);
      const propertyStructs = {};

      for (const key in schema) {
        const s = struct(schema[key]);
        propertyStructs[key] = s;
      }

      this.type = type;
      this.valueStruct = valueStruct;
      this.propertyStructs = propertyStructs;
    }

    /**
     * Get the default value for a `key`, given an initial `value`.
     *
     * @param {String} key
     * @param {Any} value
     * @return {Any}
     */

    defaultProperty(key, value) {
      if (value !== undefined) return value;
      const { defaults = {} } = this;
      const v = defaults[key];
      return v;
    }

    /**
     * Mask the properties a `value`, returning only the known ones.
     *
     * @param {Function} Struct
     * @param {Object} value
     */

    mask(value) {
      if (value === undefined) {
        return undefined;
      }

      const { schema } = this;
      const ret = {};

      for (const key in schema) {
        if (key in value) ret[key] = value[key];
      }

      return ret;
    }

    /**
     * Validate a list `element` at `index`.
     *
     * @param {Any} element
     * @param {Number} index
     * @return {Any|StructError}
     */

    validateProperty(key, value) {
      const s = this.propertyStructs[key];

      value = this.defaultProperty(key, value);

      if (!s) {
        return new _structError2.default(PROPERTY_UNKNOWN, { key, path: [key] });
      }

      const result = s.validate(value);

      if (result instanceof _structError2.default) {
        const error = result;
        const path = [key].concat(error.path);

        if (error.code === VALUE_INVALID) {
          return new _structError2.default(PROPERTY_INVALID, _extends({}, error, { key, path }));
        }

        if (error.code === VALUE_REQUIRED) {
          return new _structError2.default(PROPERTY_REQUIRED, _extends({}, error, { key, path }));
        }

        if ('path' in error) error.path = path;
        return error;
      }

      return result;
    }

    /**
     * Validate a list `value`.
     *
     * @param {Any} value
     * @return {Any|StructError}
     */

    validate(value) {
      const { options, propertyStructs, valueStruct } = this;

      value = this.default(value);

      if (options.unsealed) {
        value = this.mask(value);
      }

      const result = valueStruct.validate(value);

      if (result instanceof _structError2.default) {
        return result;
      }

      const errors = [];
      const values = {};
      const isUndefined = !options.required && value === undefined;
      let hasKeys = false;
      value = isUndefined ? {} : value;

      for (const k in value) {
        hasKeys = true;
        const v = value[k];
        const r = this.validateProperty(k, v);

        if (r instanceof _structError2.default) {
          if (r.code === PROPERTY_REQUIRED && isUndefined) {
            return new _structError2.default(VALUE_REQUIRED, { type: 'object' });
          }

          errors.push(r);
        } else {
          values[k] = r;
        }
      }

      for (const k in propertyStructs) {
        if (k in values) continue;

        const v = value[k];
        const r = this.validateProperty(k, v);

        if (r instanceof _structError2.default) {
          if (r.code === PROPERTY_REQUIRED && isUndefined) {
            return new _structError2.default(VALUE_REQUIRED, { type: 'object' });
          }

          errors.push(r);
        }
      }

      if (errors.length) {
        const first = errors[0];
        first.errors = errors;
        return first;
      }

      return isUndefined && !hasKeys ? undefined : values;
    }

  }

  /**
   * Create a `kind` struct with schema `definition`, `defaults` and `options`.
   *
   * @param {String} kind
   * @param {Function|String|Array|Object} definition
   * @param {Any} defaults
   * @param {Object} options
   * @return {Function}
   */

  function createStruct(kind, definition, defaults, options) {
    if (isStruct(definition)) {
      return definition;
    }

    const args = [definition, defaults, options];
    let schema;

    if (kind === 'function') {
      schema = new FunctionSchema(...args);
    } else if (kind === 'scalar') {
      schema = new ScalarSchema(...args);
    } else if (kind === 'list') {
      schema = new ListSchema(...args);
    } else if (kind === 'object') {
      schema = new ObjectSchema(...args);
    } else {
      throw new Error(`Unrecognized struct kind: ${kind}`);
    }

    // Define the struct creator function.
    function Struct(data) {
      if (this instanceof Struct) {
        throw new Error('The `Struct` creation function should not be used with the `new` keyword.');
      }

      return schema.assert(data);
    }

    // Add a private property for identifying struct functions.
    Struct[IS_STRUCT] = true;

    // Mix in all of the methods for this kind of schema.
    STRUCT_METHODS.forEach(method => {
      if (schema[method]) {
        Struct[method] = (...a) => schema[method](...a);
      }
    });

    return Struct;
  }

  /**
   * Define a struct with schema `definition`, `defaults` and `options`.
   *
   * @param {Function|String|Array|Object} definition
   * @param {Any} defaults
   * @param {Object} options
   * @return {Function}
   */

  function struct(definition, defaults, options) {
    if (isStruct(definition)) {
      return definition;
    }

    const kind = getKind(definition);
    const Struct = createStruct(kind, definition, defaults, options);
    return Struct;
  }

  // Mix in a factory for each kind of struct.
  STRUCT_KINDS.forEach(kind => {
    struct[kind] = (...args) => createStruct(kind, ...args);
  });

  // Mix in the convenience properties for option flags.
  STRUCT_FLAGS.forEach(flag => {
    Object.defineProperty(struct, flag, {
      get: () => (s, d, o = {}) => struct(s, d, _extends({}, o, { [flag]: true }))
    });
  });

  /**
   * Return the struct factory.
   */

  return struct;
}

/**
 * Check if a `value` is a struct.
 *
 * @param {Any} value
 * @return {Boolean}
 */

function isStruct(value) {
  return !!(value && value[IS_STRUCT]);
}

/**
 * Get the kind of a struct from its schema `definition`.
 *
 * @param {Any} definition
 * @return {String}
 */

function getKind(definition) {
  switch ((0, _kindOf2.default)(definition)) {
    case 'function':
      return 'function';
    case 'string':
      return 'scalar';
    case 'array':
      return 'list';
    case 'object':
      return 'object';
    default:
      {
        throw new Error(`A struct schema definition must be a string, array or object, but you passed: ${definition}`);
      }
  }
}

/**
 * Export.
 *
 * @type {Function}
 */

exports.default = superstruct;